/*
This source file is part of Castor3D (http://castor3d.developpez.com/castor3d.htm)

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
the program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
*/
#ifndef ___C3D_Sampler___
#define ___C3D_Sampler___

#include "Castor3DPrerequisites.hpp"

#include "Renderable.hpp"
#include "SamplerRenderer.hpp"
#include "BinaryParser.hpp"

#pragma warning( push )
#pragma warning( disable:4251 )
#pragma warning( disable:4275 )

namespace Castor3D
{
	/*!
	\author 	Sylvain DOREMUS
	\date 		25/03/2013
	\version	0.7.0
	\~english
	\brief		Defines a sampler for a texture
	\~french
	\brief		Dfinit un sampler pour une texture
	*/
	class C3D_API Sampler
		:	public Renderable< Sampler, SamplerRenderer >
	{
	public:
		/*!
		\author		Sylvain DOREMUS
		\version	0.7.0
		\date		21/03/2014
		\~english
		\brief		Sampler loader
		\remark
		\~french
		\brief		Loader de Sampler
		*/
		class C3D_API TextLoader
			:	public Castor::Loader< Sampler, Castor::eFILE_TYPE_TEXT, Castor::TextFile >
			,	public Castor::NonCopyable
		{
		public:
			/**
			 *\~english
			 *\brief		Constructor
			 *\~french
			 *\brief		Constructeur
			 */
			TextLoader( Castor::File::eENCODING_MODE p_eEncodingMode = Castor::File::eENCODING_MODE_ASCII );
			/**
			 *\~english
			 *\brief			Writes a sampler into a text file
			 *\param[in]		p_sampler	The sampler to save
			 *\param[in,out]	p_file		The file where to save the sampler
			 *\~french
			 *\brief			Ecrit un chantillonneur dans un fichier texte
			 *\param[in]		p_sampler	L'chantillonneur
			 *\param[in,out]	p_file		Le fichier
			 */
			virtual bool operator()( Sampler const & p_sampler, Castor::TextFile & p_file );
		};
		/*!
		\author		Sylvain DOREMUS
		\date		08/04/2014
		\~english
		\brief		Sampler loader
		\~english
		\brief		Loader de Sampler
		*/
		class C3D_API BinaryParser
			:	public Castor3D::BinaryParser< Sampler >
		{
		public:
			/**
			 *\~english
			 *\brief		Constructor
			 *\param[in]	p_path	The current folder path
			 *\~french
			 *\brief		Constructeur
			 *\param[in]	p_path	Le chemin d'accs au dossier courant
			 */
			BinaryParser( Castor::Path const & p_path );
			/**
			 *\~english
			 *\brief		Function used to fill the chunk from specific data
			 *\param[in]	p_obj	The object to write
			 *\param[out]	p_chunk	The chunk to fill
			 *\return		\p false if any error occured
			 *\~french
			 *\brief		Fonction utilise afin de remplir le chunk de donnes spcifiques
			 *\param[in]	p_obj	L'objet  crire
			 *\param[out]	p_chunk	Le chunk  remplir
			 *\return		\p false si une erreur quelconque est arrive
			 */
			virtual bool Fill( Sampler const & p_obj, BinaryChunk & p_chunk )const;
			/**
			 *\~english
			 *\brief		Function used to retrieve specific data from the chunk
			 *\param[out]	p_obj	The object to read
			 *\param[in]	p_chunk	The chunk containing data
			 *\return		\p false if any error occured
			 *\~french
			 *\brief		Fonction utilise afin de rcuprer des donnes spcifiques  partir d'un chunk
			 *\param[out]	p_obj	L'objet  lire
			 *\param[in]	p_chunk	Le chunk contenant les donnes
			 *\return		\p false si une erreur quelconque est arrive
			 */
			virtual bool Parse( Sampler & p_obj, BinaryChunk & p_chunk )const;
		};

	public:
		/**
		 *\~english
		 *\brief		Constructor
		 *\param[in]	p_name	    The sampler name
		 *\param[in]	p_pEngine	The core engine
		 *\~french
		 *\brief		Constructeur
		 *\param[in]	p_name	    Le nom du sampler
		 *\param[in]	p_pEngine	Le moteur
		 */
		Sampler( Engine * p_pEngine, Castor::String const & p_name = Castor::cuEmptyString );
		/**
		 *\~english
		 *\brief		Destructor
		 *\~french
		 *\brief		Destructeur
		 */
		virtual ~Sampler();
		/**
		 *\~english
		 *\brief		Initialises the sampler
		 *\return		\p true if ok
		 *\~french
		 *\brief		Initialise le sampler
		 *\return		\p true si tout s'est bien pass
		 */
		virtual bool Initialise();
		/**
		 *\~english
		 *\brief		Cleanups the sampler
		 *\~french
		 *\brief		Nettoie le sampler
		 */
		virtual void Cleanup();
		/**
		 *\~english
		 *\brief		Applies the sampler
		 *\param[in]	p_eDimension	The texture dimension
		 *\param[in]	p_uiIndex		The sampler index
		 *\return		\p true if ok
		 *\~french
		 *\brief		Applique le sampler
		 *\param[in]	p_eDimension	La dimension de la texture
		 *\param[in]	p_uiIndex		L'index du sampler
		 *\return		\p true si tout s'est bien pass
		 */
		virtual bool Bind( eTEXTURE_DIMENSION p_eDimension, uint32_t p_uiIndex );
		/**
		 *\~english
		 *\brief		Removes the sampler
		 *\~french
		 *\brief		Enlve le sampler
		 */
		virtual void Unbind();
		/**
		 *\~english
		 *\brief		Retrieves the interpolation mode for given filter
		 *\param[in]	p_eFilter	The concerned filter
		 *\return		The interpolation mode
		 *\~french
		 *\brief		Rcupre le mode d'interpolation pour le filtre donn
		 *\param[in]	p_eFilter	Le filtre concern
		 *\return		Le mode d'interpolation
		 */
		inline eINTERPOLATION_MODE GetInterpolationMode( eINTERPOLATION_FILTER p_eFilter )const
		{
			return m_eInterpolationModes[p_eFilter];
		}
		/**
		 *\~english
		 *\brief		Defines the interpolation mode for given filter
		 *\param[in]	p_eFilter	The concerned filter
		 *\param[in]	p_eMode		The wanted mode
		 *\~french
		 *\brief		Dfinit le mode d'interpolation pour le filtre donn
		 *\param[in]	p_eFilter	Le filtre concern
		 *\param[in]	p_eMode		Le mode souhait
		 */
		inline void SetInterpolationMode( eINTERPOLATION_FILTER p_eFilter, eINTERPOLATION_MODE p_eMode )
		{
			m_eInterpolationModes[p_eFilter] = p_eMode;
		}
		/**
		 *\~english
		 *\brief		Retrieves the wrapping mode for given dimension
		 *\param[in]	p_eUVW	The dimension
		 *\return		The wrapping mode
		 *\~french
		 *\brief		Rcupre le mode de wrap pour la dimension donne
		 *\param[in]	p_eUVW	La dimension
		 *\return		Le mode de wrap
		 */
		inline eWRAP_MODE GetWrappingMode( eTEXTURE_UVW p_eUVW )const
		{
			return m_eWrapModes[p_eUVW];
		}
		/**
		 *\~english
		 *\brief		Defines the wrapping mode for given dimension
		 *\param[in]	p_eUVW	The dimension
		 *\param[in]	p_eMode	The wrapping mode
		 *\~french
		 *\brief		Dfinit le mode de wrap pour la dimension donne
		 *\param[in]	p_eUVW	La dimension
		 *\param[in]	p_eMode	Le mode de wrap
		 */
		inline void SetWrappingMode( eTEXTURE_UVW p_eUVW, eWRAP_MODE p_eMode )
		{
			m_eWrapModes[p_eUVW] = p_eMode;
		}
		/**
		 *\~english
		 *\brief		Retrieves the maximal anisotropy filtering value
		 \return		The value
		 *\~french
		 *\brief		Rcupre la valeur maximale de filtrage anisotropique
		 *\return		La valeur
		 */
		inline real GetMaxAnisotropy()const
		{
			return m_rMaxAnisotropy;
		}
		/**
		 *\~english
		 *\brief		Defines the maximal anisotropy filtering value
		 *\param[in]	p_rValue	The new value
		 *\~french
		 *\brief		Dfinit la valeur maximale de filtrage anisotropique
		 *\param[in]	p_rValue	La nouvelle valeur
		 */
		inline void SetMaxAnisotropy( real p_rValue )
		{
			m_rMaxAnisotropy = p_rValue;
		}
		/**
		 *\~english
		 *\brief		Retrieves the minimal LOD level
		 \return		The level
		 *\~french
		 *\brief		Rcupre le niveau minimal pour le LOD
		 *\return		Le niveau
		 */
		inline real GetMinLod()const
		{
			return m_rMinLod;
		}
		/**
		 *\~english
		 *\brief		Defines the minimal LOD level
		 *\param[in]	p_rLod	The new value
		 *\~french
		 *\brief		Dfinit le niveau minimal pour le LOD
		 *\param[in]	p_rLod	La nouvelle valeur
		 */
		inline void SetMinLod( real p_rLod )
		{
			m_rMinLod = p_rLod;
		}
		/**
		 *\~english
		 *\brief		Retrieves the maximal LOD level
		 \return		The level
		 *\~french
		 *\brief		Rcupre le niveau maximal pour le LOD
		 *\return		Le niveau
		 */
		inline real GetMaxLod()const
		{
			return m_rMaxLod;
		}
		/**
		 *\~english
		 *\brief		Defines the maximal LOD level
		 *\param[in]	p_rLod	The new value
		 *\~french
		 *\brief		Dfinit le niveau maximal pour le LOD
		 *\param[in]	p_rLod	La nouvelle valeur
		 */
		inline void SetMaxLod( real p_rLod )
		{
			m_rMaxLod = p_rLod;
		}
		/**
		 *\~english
		 *\brief		Retrieves the LOD bias
		 \return		The level
		 *\~french
		 *\brief		Rcupre le LOD bias
		 *\return		Le niveau
		 */
		inline real GetLodBias()const
		{
			return m_rLodBias;
		}
		/**
		 *\~english
		 *\brief		Defines the LOD bias
		 *\param[in]	p_rLod	The new value
		 *\~french
		 *\brief		Dfinit le LOD bias
		 *\param[in]	p_rLod	La nouvelle valeur
		 */
		inline void SetLodBias( real p_rLod )
		{
			m_rLodBias = p_rLod;
		}
		/**
		 *\~english
		 *\brief		Retrieves the texture border colour
		 *\return		The colour
		 *\~french
		 *\brief		Rcupre la couleur de bord de la texture
		 *\return		La couleur
		 */
		inline	Castor::Colour const & GetBorderColour()const
		{
			return m_clrBorderColour;
		}
		/**
		 *\~english
		 *\brief		Defines the texture border colour
		 *\param[in]	p_crColour	The new value
		 *\~french
		 *\brief		Dfinit la couleur de bord de la texture
		 *\param[in]	p_crColour	La nouvelle valeur
		 */
		inline	void SetBorderColour( Castor::Colour const & p_crColour )
		{
			m_clrBorderColour = p_crColour;
		}
		/**
		 *\~english
		 *\brief		Retrieves the sampler name
		 \return		The name
		 *\~french
		 *\brief		Rcupre le nom de l'chantillonneur
		 *\return		Le nom
		 */
		inline Castor::String const & GetName()const
		{
			return m_name;
		}
		/**
		 *\~english
		 *\brief		Defines the sampler name
		 *\param[in]	p_name	The name
		 *\~french
		 *\brief		Rcupre le nom de l'chantillonneur
		 *\param[in]	p_name	Le nom
		 */
		inline void SetName( Castor::String const & p_name )
		{
			m_name = p_name;
		}

	private:
		//!\~english Sampler interpolation modes	\~french Modes d'interpolation du sampler
		eINTERPOLATION_MODE m_eInterpolationModes[eINTERPOLATION_FILTER_COUNT];
		//!\~english Sampler wrapping modes	\~french Modes de wrapping du sampler
		eWRAP_MODE m_eWrapModes[eTEXTURE_UVW_COUNT];
		//!\~english Minimal LOD Level	\~french Niveau de LOD minimal
		real m_rMinLod;
		//!\~english Maximal LOD Level	\~french Niveau de LOD maximal
		real m_rMaxLod;
		//!\~english The texture LOD offset	\~french Le dcalage de Lod de la texture
		real m_rLodBias;
		//!\~english Texture border colour	\~french Couleur des bords de la texture
		Castor::Colour m_clrBorderColour;
		//!\~english Maximal anisotropic filtering value	\~french Valeur maximale pour le filtrage anisotropique
		real m_rMaxAnisotropy;
		//!\~english The sampler name	\~french Le nom de l'chantillonneur
		Castor::String m_name;
	};
}

#pragma warning( pop )

#endif
