﻿/*
This source file is part of Castor3D (http://castor3d.developpez.com/castor3d.htm)

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
the program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
*/
#ifndef ___C3D_SHADER_MANAGER_H___
#define ___C3D_SHADER_MANAGER_H___

#include "Castor3DPrerequisites.hpp"

#pragma warning( push )
#pragma warning( disable:4251 )
#pragma warning( disable:4275 )

namespace Castor3D
{
	/*!
	\author 	Sylvain DOREMUS
	\date		14/02/2010
	\~english
	\brief		Manager used to hold the shader programs. Holds it, destroys it during a rendering loop
	\~french
	\brief		Manager utilisé pour garder les programmes de shaders. Il les garde et permet leur destruction au cours d'une boucle de rendu
	*/
	class C3D_API ShaderManager
	{
	public:
		/**
		 *\~english
		 *\brief		Constructor
		 *\~french
		 *\brief		Constructeur
		 */
		ShaderManager();
		/**
		 *\~english
		 *\brief		Destructor
		 *\~french
		 *\brief		Destructeur
		 */
		virtual ~ShaderManager();
		/**
		 *\~english
		 *\brief		Destroys all the shaders of the array of shaders to destroy
		 *\~french
		 *\brief		Détruit tous les shaders du tableau de shaders à détruire
		 */
		void Update();
		/**
		 *\~english
		 *\brief		Flushes the active shaders list and adds all it's shaders to the array of shaders to destroy
		 *\remark		Calls Update
		 *\~french
		 *\brief		Vide la liste des shaders actifs dans celle des shaders à détruire
		 *\remark		Appelle Update
		 */
		void Clear();
		/**
		 *\~english
		 *\brief		Flushes the active shaders list and adds all it's shaders to the array of shaders to destroy
		 *\~french
		 *\brief		Vide la liste des shaders actifs dans celle des shaders à détruire
		 */
		void ClearShaders();
		/**
		 *\~english
		 *\brief		Creates a new program
		 *\param[in]	p_eLanguage	The wanted shader language
		 *\return		The newly created program
		 *\~french
		 *\brief		Crée un nouveau programme
		 *\param[in]	p_eLanguage	Le langage de shader souhaité
		 *\return		Le programme créé
		 */
		ShaderProgramBaseSPtr GetNewProgram( eSHADER_LANGUAGE p_eLanguage = eSHADER_LANGUAGE_AUTO );
		/**
		 *\~english
		 *\brief		Looks for an automatically generated program corresponding to given flags
		 *\remark		If none exists it is created
		 *\param[in]	p_uiTextureFlags	Bitwise ORed eTEXTURE_CHANNEL
		 *\param[in]	p_uiProgramFlags	Bitwise ORed ePROGRAM_FLAG
		 *\return		The found or created program
		 *\~french
		 *\brief		Cherche un programme automatiquement généré correspondant aux flags donnés
		 *\param[in]	p_uiTextureFlags	Une combinaison de eTEXTURE_CHANNEL
		 *\param[in]	p_uiProgramFlags	Une combinaison de ePROGRAM_FLAG
		 *\return		Le programme trouvé ou créé
		 */
		ShaderProgramBaseSPtr GetAutomaticProgram( uint32_t p_uiTextureFlags, uint32_t p_uiProgramFlags );
		/**
		 *\~english
		 *\brief		Sets the RenderSystem
		 *\param[in]	p_pRenderSystem	The new value
		 *\~french
		 *\brief		Définit le RenderSystem
		 *\param[in]	p_pRenderSystem	La nouvelle valeur
		 */
		inline void SetRenderSystem( RenderSystem * const & p_pRenderSystem )
		{
			m_pRenderSystem = p_pRenderSystem;
		}

	private:
		DECLARE_MAP( uint64_t, ShaderProgramBaseWPtr, ShaderProgramWPtrUInt64 );
		//!\~english The loaded shader programs	\~french Les programmes chargés
		ShaderProgramPtrArray m_arrayPrograms;
		//!\~english The array of shaders to destroy	\~french Les programmes à détruire
		ShaderProgramPtrArray m_arrayToDelete;
		//!\~english Automatically generated shader programs, sorted by texture flags	\~french Programmes auto-générés, triés par flags de texture
		ShaderProgramWPtrUInt64Map m_mapAutogenerated;
		//!\~english The maximal number of vertices the geometry shader can output	\~french Nombre maximale de vertex qu'un shader géométrique peut sortir
		int m_nbVerticesOut;
		//!\~english The RenderSystem used to create programs	\~french Le RenderSystem utilisé pour la création des programmes
		RenderSystem * m_pRenderSystem;
	};
}

#pragma warning( pop )

#endif
