/*
This source file is part of ProceduralGenerator (https://sourceforge.net/projects/proceduralgene/)

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
the program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
*/
#ifndef ___CASTOR_PIXEL_BUFFER_BASE_H___
#define ___CASTOR_PIXEL_BUFFER_BASE_H___

#include "Pixel.hpp"
#include "Size.hpp"
#include "Position.hpp"

namespace Castor
{
	/*!
	\author		Sylvain DOREMUS
	\version	0.6.1.0
	\date		29/08/2011
	\~english
	\brief		Pixel buffer base definition
	\remark		It has 2 dimensions
	\~french
	\brief		Dfinition de la classe de base d'un buffer de Pixel
	\remark		Il a 2 dimensions
	*/
	class PxBufferBase
	{
	public:
		//!\~english Typedef over a pointer to uint8_t	\~french Typedef d'un pointeur sur uint8_t
		typedef uint8_t * iterator;
		//!\~english Typedef over a constant pointer to uint8_t	\~french Typedef d'un pointeur constant sur uint8_t
		typedef uint8_t const * const_iterator;

	public:
		/**
		 *\~english
		 *\brief		Constructor from another buffer
		 *\param[in]	p_size			Buffer dimensions
		 *\param[in]	p_ePixelFormat	Pixels format
		 *\param[in]	p_pBuffer		Data buffer
		 *\param[in]	p_eBufferFormat	Data buffer's pixels format
		 *\~french
		 *\brief		Constructeur  partir d'un autre buffer
		 *\param[in]	p_size			Dimensions du buffer
		 *\param[in]	p_ePixelFormat	Format des pixels du buffer
		 *\param[in]	p_pBuffer		Buffer de donnes
		 *\param[in]	p_eBufferFormat	Format des pixels du buffer de donnes
		 */
		PxBufferBase( Size const & p_size, ePIXEL_FORMAT p_ePixelFormat, uint8_t const * p_pBuffer = NULL, ePIXEL_FORMAT p_eBufferFormat = ePIXEL_FORMAT_A8R8G8B8 );
		/**
		 *\~english
		 *\brief		Copy Constructor
		 *\param[in]	p_pixelBuffer	The PxBufferBase object to copy
		 *\~french
		 *\brief		Constructeur par copie
		 *\param[in]	p_pixelBuffer	L'objet PxBufferBase  copier
		 */
		PxBufferBase( PxBufferBase const & p_pixelBuffer );
		/**
		 *\~english
		 *\brief		Move Constructor
		 *\param[in]	p_pixelBuffer	The PxBufferBase object to move
		 *\~french
		 *\brief		Constructeur par dplacement
		 *\param[in]	p_pixelBuffer	L'objet PxBufferBase  dplacer
		 */
		PxBufferBase( PxBufferBase && p_pixelBuffer );
		/**
		 *\~english
		 *\brief		Destructor
		 *\~french
		 *\brief		Destructeur
		 */
		virtual ~PxBufferBase();
		/**
		 *\~english
		 *\brief		Copy assignment operator
		 *\param[in]	p_pixelBuffer	The PxBufferBase object to copy
		 *\return		A reference to this PxBufferBase object
		 *\~french
		 *\brief		Oprateur d'affectation par copie
		 *\param[in]	p_pixelBuffer	L'objet PxBufferBase  copier
		 *\return		Une rfrence sur cet objet PxBufferBase
		 */
		PxBufferBase & operator =( PxBufferBase const & p_pixelBuffer );
		/**
		 *\~english
		 *\brief		Move assignment operator
		 *\param[in]	p_pixelBuffer	The PxBufferBase object to move
		 *\return		A reference to this PxBufferBase object
		 *\~french
		 *\brief		Oprateur d'affectation  par dplacement
		 *\param[in]	p_pixelBuffer	L'objet PxBufferBase  dplacer
		 *\return		Une rfrence sur cet objet PxBufferBase
		 */
		PxBufferBase & operator =( PxBufferBase && p_pixelBuffer );
		/**
		 *\~english
		 *\brief		Deletes the data buffer
		 *\~french
		 *\brief		Dtruit le buffer de donnes
		 */
		virtual void clear();
		/**
		 *\~english
		 *\brief		Initialises the data buffer to the given one
		 *\remark		Conversions are made if needed
		 *\param[in]	p_pBuffer		Data buffer
		 *\param[in]	p_eBufferFormat	Data buffer's pixels format
		 *\~french
		 *\brief		Initialise le buffer de donnes  celui donn
		 *\remark		Des conversions sont faites si besoin est
		 *\param[in]	p_pBuffer		Buffer de donnes
		 *\param[in]	p_eBufferFormat	Format des pixels du buffer de donnes
		 */
		virtual void init( uint8_t const * p_pBuffer, ePIXEL_FORMAT p_eBufferFormat );
		/**
		 *\~english
		 *\brief		Initialises the data buffer at the given size
		 *\remark		Conversions are made if needed
		 *\param[in]	p_size		Buffer dimensions
		 *\~french
		 *\brief		Initialise le buffer de donnes  la taille donne
		 *\remark		Des conversions sont faites si besoin est
		 *\param[in]	p_size		Les dimensions du buffer
		 */
		virtual void init( Size const & p_size );
		/**
		 *\~english
		 *\brief		Swaps this buffer's data with the given one's
		 *\param[in]	p_pixelBuffer	The buffer to swap
		 *\~french
		 *\brief		Echange les donnes de ce buffer avec celles du buffer donn
		 *\param[in]	p_pixelBuffer	Le buffer  changer
		 */
		virtual void swap( PxBufferBase & p_pixelBuffer );
		/**
		 *\~english
		 *\brief		Converts and assigns a data buffer to this buffer
		 *\param[in]	p_pBuffer		Data buffer
		 *\param[in]	p_eBufferFormat	Data buffer's pixels format
		 *\return
		 *\~french
		 *\brief		Convertit et assigne les donnes du buffer donn  ce buffer
		 *\param[in]	p_pBuffer		Buffer de donnes
		 *\param[in]	p_eBufferFormat	Format des pixels du buffer de donnes
		 *\return
		 */
		virtual void assign( std::vector< uint8_t > const & p_pBuffer, ePIXEL_FORMAT p_eBufferFormat ) = 0;
		/**
		 *\~english
		 *\brief		Retrieves the pointer on constant datas
		 *\return		The pointer
		 *\~french
		 *\brief		Rcupre le pointeur sur les donnes constantes
		 *\return		Les donnes
		 */
		virtual uint8_t const * const_ptr()const = 0;
		/**
		 *\~english
		 *\brief		Retrieves the pointer on datas
		 *\return		The pointer
		 *\~french
		 *\brief		Rcupre le pointeur sur les donnes
		 *\return		Les donnes
		 */
		virtual uint8_t * ptr() = 0;
		/**
		 *\~english
		 *\brief		Retrieves the total size of the buffer
		 *\return		count() * (size of a pixel)
		 *\~french
		 *\brief		Rcupre la taille totale du buffer
		 *\return		count() * (size of a pixel)
		 */
		virtual uint32_t size()const = 0;
		/**
		 *\~english
		 *\brief		Creates a new buffer with same values as this one
		 *\return		The created buffer
		 *\~french
		 *\brief		Cre un nouveau buffer avec les mmes valeurs
		 *\return		Le buffer cr
		 */
		virtual std::shared_ptr<PxBufferBase> clone()const = 0;
		/**
		 *\~english
		 *\brief		Retrieves the pixel data at given position
		 *\param[in]	x, y	The pixel position
		 *\return		The pixel data
		 *\~french
		 *\brief		Rcupre les donnes du pixel  la position donne
		 *\param[in]	x, y	The pixel position
		 *\return		Les donnes du pixel
		 */
		virtual uint8_t * get_at( uint32_t x, uint32_t y ) = 0;
		/**
		 *\~english
		 *\brief		Retrieves the pixel data at given position
		 *\param[in]	x, y	The pixel position
		 *\return		The constant pixel data
		 *\~french
		 *\brief		Rcupre les donnes du pixel  la position donne
		 *\param[in]	x, y	The pixel position
		 *\return		Les donnes constantes du pixel
		 */
		virtual uint8_t const * get_at( uint32_t x, uint32_t y )const = 0;
		/**
		 *\~english
		 *\brief		Makes a vertical swap of pixels
		 *\~french
		 *\brief		Effectue un change vertical des pixels
		 */
		virtual void flip() = 0;
		/**
		 *\~english
		 *\brief		Makes a horizontal swap of pixels
		 *\~french
		 *\brief		Effectue un change horizontal des pixels
		 */
		virtual void mirror() = 0;
		/**
		 *\~english
		 *\brief		Retrieves the pixels format
		 *\return		The pixels format
		 *\~french
		 *\brief		Rcupre le format ds pixels
		 *\return		Le format des pixels
		 */
		inline ePIXEL_FORMAT format()const
		{
			return m_ePixelFormat;
		}
		/**
		 *\~english
		 *\brief		Retrieves the buffer width
		 *\return		The buffer width
		 *\~french
		 *\brief		Rcupre la largeur du buffer
		 *\return		La largeur du buffer
		 */
		inline uint32_t width()const
		{
			return m_size.width();
		}
		/**
		 *\~english
		 *\brief		Retrieves the buffer height
		 *\return		The buffer height
		 *\~french
		 *\brief		Rcupre la hauteur du buffer
		 *\return		La hauteur du buffer
		 */
		inline uint32_t height()const
		{
			return m_size.height();
		}
		/**
		 *\~english
		 *\brief		Retrieves the buffer's dimensions
		 *\return		The buffer's dimensions
		 *\~french
		 *\brief		Rcupre les dimensions du buffer
		 *\return		Les dimensions du buffer
		 */
		inline Size const & dimensions()const
		{
			return m_size;
		}
		/**
		 *\~english
		 *\brief		Retrieves the pixels count
		 *\return		width * height
		 *\~french
		 *\brief		Rcupre le compte des pixels
		 *\return		largeur * hauteur
		 */
		inline uint32_t count()const
		{
			return width() * height();
		}
		/**
		 *\~english
		 *\brief		Retrieves the pixel data at given position
		 *\param[in]	p_position	The pixel position
		 *\return		The pixel data
		 *\~french
		 *\brief		Rcupre les donnes du pixel  la position donne
		 *\param[in]	p_position	The pixel position
		 *\return		Les donnes du pixel
		 */
		inline uint8_t * get_at( Position const & p_position )
		{
			return get_at( p_position.x(), p_position.y() );
		}
		/**
		 *\~english
		 *\brief		Retrieves the pixel data at given position
		 *\param[in]	p_position	The pixel position
		 *\return		The pixel constant data
		 *\~french
		 *\brief		Rcupre les donnes du pixel  la position donne
		 *\param[in]	p_position	The pixel position
		 *\return		Les donnes constantes du pixel
		 */
		inline uint8_t const * get_at( Position const & p_position )const
		{
			return get_at( p_position.x(), p_position.y() );
		}
		/**
		 *\~english
		 *\brief		Retrieves an iterator to the first element
		 *\return		The iterator
		 *\~french
		 *\brief		Rcupre un itrateur sur le premier lment
		 *\return		L'itrateur
		 */
		inline iterator begin()
		{
			return & m_pBuffer[0];
		}
		/**
		 *\~english
		 *\brief		Retrieves a constant iterator to the first element
		 *\return		The iterator
		 *\~french
		 *\brief		Rcupre un itrateur constant sur le premier lment
		 *\return		L'itrateur
		 */
		inline const_iterator begin()const
		{
			return & m_pBuffer[0];
		}
		/**
		 *\~english
		 *\brief		Retrieves an iterator to the last element
		 *\return		The iterator
		 *\~french
		 *\brief		Rcupre un itrateur sur le dernier lment
		 *\return		L'itrateur
		 */
		inline iterator end()
		{
			return m_pBuffer + size();
		}
		/**
		 *\~english
		 *\brief		Retrieves a constant iterator to the last element
		 *\return		The iterator
		 *\~french
		 *\brief		Rcupre un itrateur constant sur le dernier lment
		 *\return		L'itrateur
		 */
		inline const_iterator end()const
		{
			return m_pBuffer + size();
		}
		/**
		 *\~english
		 *\brief		Creates a buffer with the given data
		 *\param[in]	p_size			Buffer dimensions
		 *\param[in]	p_eWantedFormat	Pixels format
		 *\param[in]	p_pBuffer		Data buffer
		 *\param[in]	p_eBufferFormat	Data buffer's pixels format
		 *\return		The created buffer
		 *\~french
		 *\brief		Cre un buffer avec les donnes voulues
		 *\param[in]	p_size			Dimensions du buffer
		 *\param[in]	p_eWantedFormat	Format des pixels du buffer
		 *\param[in]	p_pBuffer		Buffer de donnes
		 *\param[in]	p_eBufferFormat	Format des pixels du buffer de donnes
		 *\return		Le buffer cr
		 */
		static PxBufferBaseSPtr create( Size const & p_size, ePIXEL_FORMAT p_eWantedFormat, uint8_t const * p_pBuffer = NULL, ePIXEL_FORMAT p_eBufferFormat = ePIXEL_FORMAT_A8R8G8B8 );

	private:
		ePIXEL_FORMAT m_ePixelFormat;

	protected:
		//!\~english Buffer dimensions	\~french Dimensions du buffer
		Size m_size;
		//!\~english Buffer data	\~french Donnes du buffer
		uint8_t * m_pBuffer;
	};
}

#endif
